module Eval(optimize) where

import AbsSML
import List
import PackedString
import State

optimize :: [SMLNG] -> [SMLNG]
optimize = mkTags . mkState

data StateSML = SSML State (Either PackedString WhiteSpace)

-- Remove all tags, just remember the state

mkState :: [SMLNG] -> [StateSML]
mkState = mkWork [initState]

mkWork state@(st:_) (Text s:ss) = SSML st (Left s) : mkWork state ss
mkWork state@(st:_) (Tag True tag:ss) = mkWork (newState st tag:state) ss
mkWork (st:sts) (Tag False _:ss) = mkWork sts ss
mkWork state@(st:_) (WS ws:ss) = SSML st (Right ws) : mkWork state ss
mkWork state [] = []

-- Going back to SMLNG
-- This means coming up with new tags

mkTags :: [StateSML] -> [SMLNG]
mkTags = 
  tagsWork (initState) []
  where tagsWork state stack (SSML s e:ss)
	    = tags ++ mkSML e : tagsWork (fixState stack state tags) stack' ss
              where (tags,stack') = miniTags (isWS e) state s stack
		    isWS (Left _) = False
		    isWS (Right _) = True
        tagsWork _ stack [] = clearStack stack

fixState (st:sts) state (Tag False _:ss) = fixState sts (snd st) ss
fixState stack state (Tag True tag:ss)
 = fixState ((tag,state):stack) (newState state tag) ss
fixState _ state [] = state

-- This function computes the smallest (well..) amount of tags required to
-- go from one state to another
miniTags :: Bool -> State -> State -> [(TagSort,State)] -> ([SMLNG],[(TagSort,State)])
miniTags white currentState reachState stack
    = head $ sortBy comp $
      concatMap (patch white reachState) (candidates stack currentState)

comp :: ([SMLNG],[(TagSort,State)]) -> ([SMLNG],[(TagSort,State)]) -> Ordering
-- No PL shortcut, some color nesting
{-
comp (sml1,stack1) (sml2,stack2)
    = compare (cost' sml1 + stackcost (drop (length sml1) stack1))
              (cost' sml2 + stackcost (drop (length sml2) stack2))
-}

-- No PL shortcut, no crazy PL shortcut
{-
comp (sml1,stack1) (sml2,stack2)
    = compare (cost' sml1 + stackcost stack1)
              (cost' sml2 + stackcost stack2)
-}

-- Crazy PL shortcut, No PL shortcut
comp (sml1,stack1) (sml2,stack2)
    = compare (cost' sml1 + (stackcost stack1 `div` 2))
              (cost' sml2 + (stackcost stack2 `div` 2))


-- No color nesting, no overlap inversion, crazy PL shortcut
{-
comp (sml1,stack1) (sml2,stack2)
    = compare (cost' sml1)
              (cost' sml2)
-}
-- (current state,current stack,closing tags,opening tags)
candidates stack state = firstCandidates ++ withPL
  where firstCandidates = (state,stack,[],[]) : unfold [] stack

        withPL = map (\ (state,stack,close,open) ->
		 (newState state Plaintext,
                  (Plaintext,state):stack,close,Plaintext:open) )
                 firstCandidates

        unfold foo [] = []
	unfold foo ((tag,state):ss) = (state,ss,newfoo,[]) : unfold (newfoo) ss
	  where newfoo = foo ++ [tag]

patch :: Bool -> State -> (State,[(TagSort,State)],[TagSort],[TagSort])
      -> [([SMLNG],[(TagSort,State)])]
patch iswhite state cand
    = case mkTag iswhite state cand of
         Right sol -> [sol]
	 Left cands -> concatMap (patch iswhite state) cands

data Diff
  = B Bool
  | Em Bool
  | I Bool
  | S Bool
  | TTD Bool
  | U Int
  | SizeD Int
  | Col Color
    deriving Show

mkTag iswhite state (st,stack,close,open)
    = let diffs = stateDiff st state in 
      let diffs' = if iswhite then filter isWSdiff diffs else diffs in
      case diffs' of
       [] -> Right (mkOCTags close open,stack)
       (B True:_) -> nope Bold st stack close open
       (Em True:_) -> nope Emphasis st stack close open
       (Em False:_) -> nope Emphasis st stack close open
       (I True:_) -> nope Italic st stack close open
       (S True:_) -> nope Strong st stack close open
       (TTD True:_) -> nope TT st stack close open
       (U _:_) -> nope Underline st stack close open
       (SizeD i:_) -> if i >= 0 then nope (Size i) st stack close open
		      else Left []
       (Col c:_) -> if c /= Dunno then nope (Color c) st stack close open
		    else Left []
       _ -> Left [] --Dangerous!!!

  where mkOCTags close open = map (Tag False) close ++ map (Tag True) open

	nope cons st stack close open
	    = Left [(newState st cons,(cons,st):stack,close,open++[cons])]

isWSdiff (U _) = True
isWSdiff (Col _) = True
isWSdiff (TTD _) = True
isWSdiff _ = False

stateDiff :: State -> State -> [Diff]
stateDiff s1 s2 = add s1 s2 b B $
		  add s1 s2 em Em $
		  add s1 s2 i I $
		  add s1 s2 s S $
		  add s1 s2 tt TTD $
		  add s1 s2 u U $
		  if (color s1) == (color s2) then [] else [Col (color s2)]

add s1 s2 sel cons = if sel s1 /= sel s2 then ((cons (sel s2)):) else id


mkSML (Left s) = Text s
mkSML (Right ws) = WS ws

clearStack [] = []
clearStack ((s,_):ss) = Tag False s : clearStack ss

stackcost = sum . map (cost . fst)

cost' = sum . map smlcost
  where smlcost (Tag _ sort) = cost sort
	smlcost _ = 0

